#!/usr/bin/env python
#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

SQLSTATE_CONNECTION_WAS_NOT_ESTABLISHED = "08001"
SQLSTATE_CONNECTION_ALREADY_EXISTS = "08002"
SQLSTATE_CONNECTION_NOT_EXISTS = "08003"
SQLSTATE_CONNECTION_REJECTED = "08004"
SQLSTATE_CONNECTION_FAILED_BUT_REESTABLISHED = "08506"
SQLSTATE_HOST_NOT_FOUND = "08508"
SQLSTATE_FEATURE_NOT_SUPPORTED = "0A000"
SQLSTATE_IO_ERROR = "58030"
